import typing as T
from copy import copy
from ctypes import ArgumentError
import functools

from rich.console import Console
from rich.table import Table


from .arg import Empty, Arg, get_func_argobjs


console = Console()


class ArgsCheckError(Exception):
    pass


def parse_pass_in(
        args: tuple, kwargs: dict,
        arg_objs: T.OrderedDict[str, "Arg"]):
    """Get the pass in value of the func
    arguments according to it's signature."""
    args_ = list(args)
    kwargs = copy(kwargs)
    res = {}
    for n, a in arg_objs.items():
        has_default = a.default is not Empty
        if len(args_) > 0:
            res[n] = args_.pop(0)
        elif (len(kwargs) > 0) and (n in kwargs):
            res[n] = kwargs.pop(n)
        else:
            if has_default:
                res[n] = a.default
            else:
                raise ArgumentError(
                    f"{n} is not provided and has no default value.")
    return res


def one(func=None, **kwargs):
    if func is None:
        return functools.partial(one, **kwargs)
    return One(func, **kwargs)


class One(object):
    def __init__(self, func, print_args=True, name=None):
        self.func = func
        functools.update_wrapper(self, func)
        self.arg_objs = get_func_argobjs(func)
        if name is not None:
            self.name = name
        elif hasattr(func, "name"):
            self.name = func.name
        else:
            self.name = func.__name__
        self.is_print_args = print_args
        self.table = None

    def __call__(self, *args, **kwargs):
        if self.is_print_args:
            self.table = self.get_argument_table()
        # check args
        vals = parse_pass_in(args, kwargs, self.arg_objs)
        errors = []
        for n, arg in self.arg_objs.items():
            self._check_arg(n, arg, vals[n], errors)
        if self.is_print_args:
            self.print_args()
        if len(errors) > 0:
            raise ArgsCheckError(errors)
        return self.func(*args, **kwargs)

    def print_args(self):
        if self.name:
            console.print(f"Run: [bold purple]{self.name}")
        console.print("Arguments table:\n")
        console.print(self.table)
        console.print()

    def __get__(self, obj, objtype):
        """Support instance method
        see https://stackoverflow.com/a/3296318/8500469"""
        return functools.partial(self.__call__, obj)

    def _check_arg(
            self, name: str, arg: Arg,
            val: T.Any, errors: T.List[Exception]):
        val_str = str(val)
        range_str = str(arg.range)
        tp_str = str(type(val))
        ann_tp_str = str(arg.type)
        try:
            arg.check(val)
        except Exception as e:
            errors.append(e)
            if isinstance(e, ValueError):
                val_str = f"[red]{val_str}[/red]"
                range_str = f"[red]{range_str}[/red]"
            elif isinstance(e, TypeError):
                ann_tp_str = f"[red]{ann_tp_str}[/red]"
                tp_str = f"[red]{tp_str}[/red]"
            else:
                raise e
        if self.is_print_args:
            self.table.add_row(name, ann_tp_str, range_str, val_str, tp_str)

    @staticmethod
    def get_argument_table():
        table = Table(
            show_header=True, header_style="bold magenta",
            box=None)
        table.add_column("Argument")
        table.add_column("Type")
        table.add_column("Range")
        table.add_column("InputVal")
        table.add_column("InputType")
        return table

    def cli(self):
        from fire import Fire
        Fire(self.__call__)

    def qt_gui(self, **kwargs):
        from .qt import GUI
        return GUI(self, **kwargs)()

    def dash_app(self, **kwargs):
        from .dash_app import App
        return App(self, **kwargs)()



import sys
import shlex
import typing as T
import inspect
import re
import subprocess as subp
from collections import OrderedDict
from ctypes import ArgumentError
from threading import Thread
from queue import Queue

import yaml

from ..arg import Arg, Empty
from ..core import parse_pass_in


def load_config(path: str) -> dict:
    with open(path) as f:
        conf = yaml.safe_load(f)
    return conf


def extrace_key(d: dict, key, default) -> T.Any:
    if key in d:
        return d.pop(key)
    else:
        return default


def parse_arg_objs(config: dict) -> T.OrderedDict[str, Arg]:
    args = OrderedDict()
    args_conf: dict = config['arguments']
    for n, p_conf in args_conf.items():
        p_conf: dict = p_conf.copy()
        _tp = extrace_key(p_conf, 'type', 'str')
        _default = extrace_key(p_conf, 'default', Empty)
        _range = extrace_key(p_conf, 'range', None)
        arg = Arg(
            type=eval(_tp),
            range=_range,
            default=_default,
            **p_conf
        )
        args[n] = arg
    return args


def compose_signature(args: T.OrderedDict[str, Arg]) -> inspect.Signature:
    parameters = []
    for name, arg_obj in args.items():
        param = inspect.Parameter(
            name, inspect.Parameter.POSITIONAL_OR_KEYWORD,
            default=arg_obj.default, annotation=arg_obj)
        parameters.append(param)
    sig = inspect.Signature(parameters)
    return sig


class Command(object):
    def __init__(self, template: str):
        self.template = template
        self.placeholders = [
            p.strip("{}") for p in
            re.findall(r"\{.*?\}", self.template)
        ]

    def check_placeholder(self, arg_names: T.List[str]):
        for arg in arg_names:
            if arg not in self.placeholders:
                raise ArgumentError(
                    f"The argument {arg} is not in command templates.")

    def format(self, vals: dict):
        for ph in self.placeholders:
            if ph not in vals:
                raise ValueError(
                    f"The value of placeholder {ph} is not provided.")
        cmd = self.template.format(**vals)
        return cmd


def replace_vals(vals: dict, arg_objs: T.OrderedDict[str, Arg]) -> dict:
    vals = vals.copy()
    for key, val in vals.items():
        arg_obj = arg_objs[key]
        if (val is True) and ('true_insert' in arg_obj.kwargs):
            vals[key] = arg_obj.kwargs['true_insert']
        if (val is False) and ('false_insert' in arg_obj.kwargs):
            vals[key] = arg_obj.kwargs['false_insert']
    return vals


def run_process(exe):
    # https://stackoverflow.com/a/4760274/8500469
    p = subp.Popen(exe, stdout=subp.PIPE, stderr=subp.PIPE)
    while True:
        # returns None while subprocess is running
        retcode = p.poll()
        out_line = p.stdout.readline()
        if out_line:
            yield 'stdout', out_line.decode()
        err_line = p.stderr.readline()
        if err_line:
            yield 'stderr', err_line.decode()
        if retcode is not None:
            break
    return retcode


class ProcessRunner(object):
    """Subprocess runner, allow stream stdout and stderr."""
    def __init__(self, command: str) -> None:
        self.command = command
        self.queue = Queue()
        self.proc = None
        self.t_stdout = None
        self.t_stderr = None

    def run(self):
        exe = shlex.split(self.command)
        self.proc = subp.Popen(exe, stdout=subp.PIPE, stderr=subp.PIPE)
        self.proc.stdout
        self.t_stdout = Thread(
            target=self.reader_func, args=(self.proc.stdout, self.queue))
        self.t_stdout.start()
        self.t_stderr = Thread(
            target=self.reader_func, args=(self.proc.stderr, self.queue))
        self.t_stderr.start()

    @staticmethod
    def reader_func(pipe: T.IO[bytes], queue: "Queue"):
        """https://stackoverflow.com/a/31867499/8500469"""
        try:
            with pipe:
                for line in iter(pipe.readline, b''):
                    queue.put((pipe, line))
        finally:
            queue.put(None)

    def stream(self):
        """https://stackoverflow.com/a/31867499/8500469"""
        for _ in range(2):
            for source, line in iter(self.queue.get, None):
                if source is self.proc.stdout:
                    src = "stdout"
                else:
                    src = "stderr"
                yield src, line.decode()
        return self.proc.wait()


class WrapCLI(object):
    def __init__(self, config: dict, print_cmd=True):
        self.config = config
        self.name = self.config['name']
        self.arg_objs = parse_arg_objs(config)
        self.command = Command(config['command'])
        self.command.check_placeholder(list(self.arg_objs.keys()))
        self.__signature__ = compose_signature(self.arg_objs)
        self.is_print_cmd = print_cmd

    def __call__(self, *args, **kwargs) -> int:
        vals = parse_pass_in(args, kwargs, self.arg_objs)
        vals = replace_vals(vals, self.arg_objs)
        cmd_str = self.command.format(vals)
        if self.is_print_cmd:
            print(f"Run command: {cmd_str}")
        runner = ProcessRunner(cmd_str)
        runner.run()
        g = runner.stream()
        retcode = None
        while True:
            try:
                src, line = next(g)
                if src == 'stdout':
                    print(line.rstrip("\n"))
                else:
                    print(line.rsplit("\n"), file=sys.stderr)
            except StopIteration as e:
                retcode = e.value
                break
        return retcode

    @staticmethod
    def from_config_file(path: str, *args, **kwargs) -> "WrapCLI":
        return WrapCLI(load_config(path), *args, **kwargs)



import typing as T
import inspect
from collections import OrderedDict


class Empty:
    pass


class ArgMeta(type):
    def __getitem__(self, args):
        if isinstance(args, tuple):
            return Arg(*args)
        else:
            return Arg(args)


class Arg(metaclass=ArgMeta):
    type_to_range_checker = {}
    type_to_type_checker = {}

    def __init__(self, type: type = str, range=None, default=Empty, **kwargs):
        self.type = type
        self.range = range
        self.default = default
        self.kwargs = kwargs

    @property
    def type_checker(self):
        return self.type_to_type_checker.get(self.type.__name__, None)

    @property
    def range_checker(self):
        return self.type_to_range_checker.get(self.type.__name__, None)

    def check(self, val):
        if (self.range_checker is None) and (self.type_checker is None):
            raise NotImplementedError(
                f"Not checker registered for type: {type}")
        if self.type_checker is not None:
            if not self.type_checker(val, self.type):
                raise TypeError(
                    f"Input value {val} is not in valid type({self.type})")
        if (self.range_checker is not None):
            if (not self.range_checker(val, self.range)):
                raise ValueError(f"Input value {val} is not in a valid range.")

    @classmethod
    def register_range_check(cls, type, checker):
        cls.type_to_range_checker[type.__name__] = checker

    @classmethod
    def register_type_check(cls, type, checker=None):
        checker = checker or (lambda v, t: isinstance(v, t))
        cls.type_to_type_checker[type.__name__] = checker


# register basic types
def _check_number_in_range(v, range):
    return (range is None) or (range[0] <= v <= range[1])


Arg.register_type_check(Empty, lambda v, range: True)
Arg.register_type_check(str)
Arg.register_range_check(int, _check_number_in_range)
Arg.register_type_check(int)
Arg.register_range_check(float, _check_number_in_range)
Arg.register_type_check(float)
Arg.register_type_check(bool)


def get_func_argobjs(func: T.Callable) -> T.OrderedDict[str, Arg]:
    args = OrderedDict()
    sig = inspect.signature(func)
    for n, p in sig.parameters.items():
        ann = p.annotation
        if isinstance(ann, Arg):
            arg = ann
        elif isinstance(ann, tuple):
            arg = Arg(*ann)
        elif ann is inspect._empty:
            arg = Arg(Empty)
        else:
            arg = Arg(ann)
        if p.default is not inspect._empty:
            arg.default = p.default
        args[n] = arg
    return args



import functools
from qtpy import QtWidgets
from qtpy import QtCore

from .types import (InputPath, OutputPath, Selection, SubSet)
from .arg import Empty, get_func_argobjs


class Worker(QtCore.QObject):
    finished = QtCore.Signal()

    def __init__(self, func, func_kwargs, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.func = func
        self.func_kwargs = func_kwargs
        self.result = None

    def run(self):
        self.result = self.func(**self.func_kwargs)
        self.finished.emit()


def gui(func=None, **kwargs):
    if func is None:
        return functools.partial(gui, **kwargs)
    return GUI(func, **kwargs)


class GUI():

    type_to_widget_constructor = {}

    def __init__(self, func, name=None, size=None,
                 run_once=True):
        self.func = func
        self.run_once = run_once
        self.result = None
        self.app = QtWidgets.QApplication([])
        self.main_window = QtWidgets.QMainWindow()
        if name is not None:
            name = name
        elif hasattr(func, "name"):
            name = func.name
        else:
            name = func.__name__
        self.main_window.setWindowTitle(name)
        if size:
            self.main_window.setFixedSize(*size)
        self.arg_widgets = {}
        self.window = QtWidgets.QWidget()
        self.layout = QtWidgets.QVBoxLayout()
        self.compose_ui()
        self.connect_events()

    def compose_ui(self):
        self.compose_arg_widgets(self.layout)
        self.run_btn = QtWidgets.QPushButton("Run")
        self.layout.addWidget(self.run_btn)
        self.terminal = QtWidgets.QTextEdit()
        self.window.setLayout(self.layout)
        self.main_window.setCentralWidget(self.window)

    def compose_arg_widgets(self, layout: QtWidgets.QVBoxLayout):
        arg_objs = get_func_argobjs(self.func)
        for n, a in arg_objs.items():
            if a.type is Empty:
                continue
            if a.type.__name__ not in self.type_to_widget_constructor:
                raise NotImplementedError(
                  f"Input widget constructor is not registered for {a.type}")
            constructor = self.type_to_widget_constructor[a.type.__name__]
            default = None if a.default is Empty else a.default
            w = constructor(n, a.range, default, attrs=a.kwargs)
            self.arg_widgets[n] = w
            layout.addWidget(w)

    def connect_events(self):
        self.run_btn.clicked.connect(self.run_func)

    def get_args(self):
        kwargs = {}
        for n, w in self.arg_widgets.items():
            kwargs[n] = w.get_value()
        return kwargs

    def run_func(self):
        kwargs = self.get_args()
        if self.run_once:
            self.main_window.hide()
            self.result = self.func(**kwargs)
            self.main_window.close()
        else:
            thread = self.thread = QtCore.QThread()
            worker = self.worker = Worker(self.func, kwargs)
            worker.moveToThread(thread)
            thread.started.connect(worker.run)
            worker.finished.connect(thread.quit)
            worker.finished.connect(worker.deleteLater)
            thread.finished.connect(thread.deleteLater)
            thread.start()
            self.run_btn.setEnabled(False)

            def finish():
                self.result = worker.result
                self.run_btn.setEnabled(True)
            thread.finished.connect(finish)

    def __call__(self):
        self.main_window.show()
        self.app.exec()
        return self.result

    @classmethod
    def register_widget(cls, type_, widget_constructor):
        cls.type_to_widget_constructor[type_.__name__] = widget_constructor


class InputItem(QtWidgets.QWidget):
    def __init__(self, name, range, default, attrs, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.range = range
        self.default = default
        self.attrs = attrs
        self.init_layout()
        self.init_ui()
        self.setLayout(self.layout)

    def init_layout(self):
        self.layout = QtWidgets.QHBoxLayout()

    def init_ui(self, label_stretch=1):
        label = self.attrs.get('text', self.name)
        self.label = QtWidgets.QLabel(f"{label}:")
        if label_stretch:
            self.layout.addWidget(self.label, stretch=label_stretch)
        else:
            self.layout.addWidget(self.label)

    def get_value(self):
        return self.input.value()


class IntInputItem(InputItem):
    def init_ui(self):
        super().init_ui()
        self.input = QtWidgets.QSpinBox()
        if self.range:
            self.input.setMinimum(self.range[0])
            self.input.setMaximum(self.range[1])
        if self.default:
            self.input.setValue(self.default)
        self.layout.addWidget(self.input, stretch=1)


class FloatInputItem(InputItem):
    def init_ui(self):
        super().init_ui()
        self.input = QtWidgets.QDoubleSpinBox()
        self.input.setSingleStep(0.1)
        if self.range:
            self.input.setMinimum(self.range[0])
            self.input.setMaximum(self.range[1])
        if self.default:
            self.input.setValue(self.default)
        self.layout.addWidget(self.input, stretch=1)


class StrInputItem(InputItem):
    def init_ui(self):
        super().init_ui()
        self.input = QtWidgets.QLineEdit()
        if self.default:
            self.input.setText(self.default)
        self.layout.addWidget(self.input, stretch=1)

    def get_value(self):
        return self.input.text()


class BoolInputItem(InputItem):
    def init_ui(self):
        super().init_ui(label_stretch=2)
        self.bt = QtWidgets.QRadioButton("True")
        self.bf = QtWidgets.QRadioButton("False")
        self.layout.addWidget(self.bt, stretch=1)
        self.layout.addWidget(self.bf, stretch=1)
        if (self.default is False):
            self.bf.setChecked(True)
        else:
            self.bt.setChecked(True)

    def get_value(self):
        return self.bt.isChecked()


class SelectionInputItem(InputItem):
    def init_ui(self):
        super().init_ui()
        self.input = QtWidgets.QComboBox()
        if self.range:
            self.input.addItems(self.range)
        if self.default:
            self.input.setCurrentText(self.default)
        self.layout.addWidget(self.input, stretch=1)

    def get_value(self):
        return self.input.currentText()


class SubsetInputItem(InputItem):
    def init_layout(self):
        self.layout = QtWidgets.QVBoxLayout()

    def init_ui(self):
        super().init_ui(label_stretch=None)
        self.cb_layout = QtWidgets.QHBoxLayout()
        self.cbs = []
        for it in self.range:
            cb = QtWidgets.QCheckBox(it)
            self.cb_layout.addWidget(cb)
            self.cbs.append(cb)
        self.layout.addLayout(self.cb_layout)
        if self.default:
            for val in self.default:
                self.cbs[self.range.index(val)].setChecked(True)

    def get_value(self):
        res = []
        for i, val in enumerate(self.range):
            if self.cbs[i].isChecked():
                res.append(val)
        return res


class PathInputItem(InputItem):
    def init_layout(self):
        self.layout = QtWidgets.QVBoxLayout()

    def init_ui(self):
        super().init_ui(label_stretch=None)
        self.box = QtWidgets.QHBoxLayout()
        self.path_edit = QtWidgets.QLineEdit()
        self.dialog_open = QtWidgets.QPushButton("open")
        self.dialog_open.clicked.connect(self.get_path)
        self.box.addWidget(self.path_edit, stretch=2)
        if self.default:
            self.path_edit.setText(self.default)
        self.box.addWidget(self.dialog_open, stretch=1)
        self.layout.addLayout(self.box)

    def get_path(self):
        fname, _ = QtWidgets.QFileDialog.getOpenFileName(
            None, "Open file", "", "All files (*)")
        self.path_edit.setText(fname)

    def get_value(self):
        return self.path_edit.text()


class OutPathInputItem(PathInputItem):
    def get_path(self):
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            None, "Save file", "", "All files (*)")
        self.path_edit.setText(fname)


GUI.register_widget(int, IntInputItem)
GUI.register_widget(float, FloatInputItem)
GUI.register_widget(str, StrInputItem)
GUI.register_widget(bool, BoolInputItem)
GUI.register_widget(Selection, SelectionInputItem)
GUI.register_widget(SubSet, SubsetInputItem)
GUI.register_widget(InputPath, PathInputItem)
GUI.register_widget(OutputPath, OutPathInputItem)



from pathlib import Path

from oneface.core import Arg


class ArgType(object):
    @staticmethod
    def check_type(val, type_):
        return True

    @staticmethod
    def check_range(val, range_):
        return True


class Selection(ArgType):
    @staticmethod
    def check_range(val, range_):
        return val in range_


class SubSet(ArgType):
    @staticmethod
    def check_range(val, range_):
        return all([v in range_ for v in val])


class InputPath(ArgType):
    @staticmethod
    def check_type(val, type_):
        return isinstance(val, str) or isinstance(val, Path)

    @staticmethod
    def check_range(val, range_):
        path = Path(val) if isinstance(val, str) else val
        return path.exists()


class OutputPath(ArgType):
    @staticmethod
    def check_type(val, type_):
        return isinstance(val, str) or isinstance(val, Path)


Arg.register_range_check(Selection, Selection.check_range)
Arg.register_range_check(SubSet, SubSet.check_range)
Arg.register_type_check(InputPath, InputPath.check_type)
Arg.register_range_check(InputPath, InputPath.check_range)
Arg.register_type_check(OutputPath, OutputPath.check_type)



import sys
import typing as T
import os.path as osp

import fire

from .wrap import WrapCLI
from ..core import one


FILE_DIR = osp.dirname(osp.abspath(__file__))
EXAMPLE_FILE = osp.join(FILE_DIR, "example.yaml")


def gen(target_path: str = "./example.yaml"):
    print(f"Generate an example config file in: {target_path}")
    with open(EXAMPLE_FILE) as f:
        content = f.read()
    with open(target_path, 'w') as f:
        f.write(content)


InterfaceTypes = T.Literal["cli", "qt_gui", "dash_app"]


def run(
        config_path: str, interface: InterfaceTypes,
        print_cmd: bool = True, **kwargs):
    """
    :param config_path: The path to your config(.yaml) file.
    :param interface: The interface type, 'qt_gui' | 'dash_app' | 'cli'
    :param print_cmd: Print the actually executed command or not.
    """
    wrap = WrapCLI.from_config_file(config_path, print_cmd=print_cmd)
    config = wrap.config
    of = one(wrap, **kwargs)
    if interface == "qt_gui":
        ret_code = of.qt_gui(**config.get('qt_config', {}))
    elif interface == "dash_app":
        ret_code = of.dash_app(**config.get('dash_config', {}))
    else:
        ret_code = of.cli()
    sys.exit(ret_code)


if __name__ == "__main__":
    fire.Fire({
        'run': run,
        'generate': gen,
    })



import contextlib
import functools
from io import StringIO
import os.path as osp

from dash import Dash, dcc, html, Input, Output, State
from dash.exceptions import PreventUpdate
from ansi2html import Ansi2HTMLConverter
import visdcc

from .types import (Selection, SubSet, InputPath, OutputPath)
from .arg import Empty, get_func_argobjs


HERE = osp.abspath(osp.dirname(__file__))


def app(func=None, **kwargs):
    if func is None:
        return functools.partial(app, **kwargs)
    return App(func, **kwargs)


class App(object):
    type_to_widget_constructor = {}
    convert_types = {}

    def __init__(
            self, func, name=None,
            show_console=True,
            console_interval=2000,
            interactive=False, init_run=False,
            result_show_type="text",
            **server_args):
        self.func = func
        if name is not None:
            self.name = name
        elif hasattr(func, "name"):
            self.name = func.name
        else:
            self.name = func.__name__
        self.show_console = show_console
        self.console_interval = console_interval
        self.interactive = interactive
        self.init_run = init_run
        self.result_show_type = result_show_type
        self.server_args = server_args
        self.input_names = None
        self.input_types = None
        self.input_attrs = None
        self.result = None
        self.dash_app = self.get_dash_app()

    def get_layout(self):
        widgets, names, types, attrs = self.parse_args()
        self.input_names = names
        self.input_types = types
        self.input_attrs = attrs
        sub_nodes = [
            html.H3("Arguments"),
            *widgets,
            html.Br(),
            html.Button("Run", id="run-btn"),
            html.Div("", style={"height": "20px"}),
        ]
        if self.show_console:
            sub_nodes += self.get_console_layout()
        sub_nodes += self.get_result_layout()
        layout = html.Div(children=sub_nodes, style={
            'width': "60%",
            'min-width': "400px",
            'max-width': "800px",
            'margin': "auto",
        })
        return layout

    def get_console_layout(self):
        return [
            html.H3("Console"),
            html.Div("", style={"height": "20px"}),
            dcc.Interval(
                id="console-interval",
                interval=self.console_interval, n_intervals=0),
            html.Iframe(id="console-out", style={
                "width": "100%",
                "max-width": "100%",
                "height": "400px",
                "resize": "both"
            }),
            visdcc.Run_js(id="jsscroll", run="")
        ]

    def get_result_layout(self):
        layout = [
            html.H3("Result"),
        ]
        if self.result_show_type == "text":
            layout += [
                html.Div("", id="out"),
            ]
        elif self.result_show_type == "download":
            layout += [
                html.Div("", id="out", style={'display': 'none'}),
                html.Button("Download Result", id="res-download-btn"),
                dcc.Download(id="res-download-index")
            ]
        return layout

    def parse_args(self):
        widgets, names, types, attrs = [], [], [], []
        arg_objs = get_func_argobjs(self.func)
        for n, a in arg_objs.items():
            if a.type is Empty:
                continue
            constructor = self.type_to_widget_constructor[a.type.__name__]
            default = None if a.default is Empty else a.default
            attr = a.kwargs
            widget = constructor(
                n, a.range, default, attrs=attr).widget
            widgets.append(widget)
            names.append(n)
            types.append(a.type)
            attrs.append(attr)
        return widgets, names, types, attrs

    def get_run_callback_decorator(self, app):
        inputs = [Input("run-btn", 'n_clicks')]
        for i, n in enumerate(self.input_names):
            is_interactive = (
                self.interactive or
                (self.input_attrs[i].get('interactive') is True)
            )
            id_ = f"input-{n}"
            if is_interactive:
                input = Input(id_, "value")
            else:
                input = State(id_, "value")
            inputs.append(input)
        output = Output("out", "children")
        deco = app.callback(output, *inputs)
        return deco

    def get_dash_app(self, *args, **kwargs):
        name = self.name or self.func.__name__
        css = ['https://codepen.io/chriddyp/pen/bWLwgP.css']
        if 'external_stylesheets' not in kwargs:
            kwargs['external_stylesheets'] = css
        app = Dash(name, *args, **kwargs)
        app.layout = self.get_layout()
        console_buffer = StringIO()

        @self.get_run_callback_decorator(app)
        def run(n_clicks, *args):
            if (not self.init_run) and (n_clicks is None):
                raise PreventUpdate
            kwargs = dict(zip(self.input_names, args))
            for i, (k, v) in enumerate(kwargs.items()):
                input_type = self.input_types[i]
                tp_name = input_type.__name__
                if tp_name in self.convert_types:
                    kwargs[k] = self.convert_types[tp_name](v)
            with contextlib.redirect_stdout(console_buffer), \
                 contextlib.redirect_stderr(console_buffer):
                self.result = self.func(**kwargs)
            return self.result

        if self.show_console:
            self.add_console_callbacks(app, console_buffer)

        if self.result_show_type == "download":
            self.add_download_callbacks(app)

        return app

    def add_console_callbacks(self, app, console_buffer):
        @app.callback(
            Output("console-out", "srcDoc"),
            Input("console-interval", "n_intervals"))
        def update_console(n):
            conv = Ansi2HTMLConverter()
            console_buffer.seek(0)
            lines = console_buffer.readlines()
            html_ = conv.convert("".join(lines))
            return html_

        doc_cache = None

        @app.callback(
            Output('jsscroll', 'run'),
            Input('console-out', 'srcDoc'))
        def scroll(doc):
            scroll_cmd = """
            var out = document.getElementById('console-out');
            out.contentWindow.scrollTo(0, 999999999);
            """
            nonlocal doc_cache
            if doc == doc_cache:
                cmd = ""
            else:
                cmd = scroll_cmd
            doc_cache = doc
            return cmd

    def add_download_callbacks(self, app):
        @app.callback(
            Output("res-download-index", "data"),
            Input("res-download-btn", "n_clicks"),
            prevent_initial_call=True)
        def send_file(n_clicks):
            return dcc.send_file(self.result)

    @classmethod
    def register_widget(cls, type, widget_constructor):
        cls.type_to_widget_constructor[type.__name__] = widget_constructor

    @classmethod
    def register_type_convert(cls, type, converter=None):
        if converter is None:
            converter = type
        cls.convert_types[type.__name__] = converter

    def __call__(self):
        self.dash_app.run_server(**self.server_args)


class InputItem(object):
    def __init__(self, name, range, default, attrs=None):
        self.name = name
        self.range = range
        self.default = default
        self.attrs = attrs
        self.input = self.get_input()
        self.input.id = f"input-{name}"
        self.widget = self.get_widget()

    def get_input(self):
        pass

    def get_widget(self):
        label = self.attrs.get("text", self.name)
        return html.Div([
            html.Div(f"{label}: ", style={
                "margin-top": "10px",
                "font-size": "20px",
            }),
            self.input
        ])


class IntInputItem(InputItem):
    def get_input(self):
        _range = self.range or [0, 100]
        _default = _range[0] if (self.default is None) else self.default
        return dcc.Input(
            min=_range[0], max=_range[1], type="number", step=1,
            value=_default, style={
                'width': "100%",
            }
        )


class FloatInputItem(InputItem):
    def get_input(self):
        _range = self.range or [0, 100]
        _default = _range[0] if (self.default is None) else self.default
        return dcc.Slider(
            _range[0], _range[1], step=None,
            value=_default,
        )


class StrInputItem(InputItem):
    def get_input(self):
        return dcc.Input(
            placeholder="Enter a value...",
            type="text",
            value=(self.default or ""),
            style={
                "width": "100%",
                "height": "40px",
                "margin": "5px",
                "font-size": "20px",
            }
        )


class BoolInputItem(InputItem):
    def get_input(self):
        _default = True if self.default is None else self.default
        return dcc.RadioItems(
            ["True", "False"],
            value=str(_default)
        )


class DropdownInputItem(InputItem):
    def get_input(self):
        return dcc.Dropdown(
            self.range,
            value=(self.default or self.range[0])
        )


class MultiDropdownInputItem(InputItem):
    def get_input(self):
        return dcc.Dropdown(
            self.range,
            value=(self.default or self.range[0]), multi=True
        )


App.register_widget(int, IntInputItem)
App.register_widget(float, FloatInputItem)
App.register_type_convert(float)
App.register_widget(str, StrInputItem)
App.register_widget(bool, BoolInputItem)
App.register_type_convert(bool, lambda s: s == "True")
App.register_widget(Selection, DropdownInputItem)
App.register_widget(SubSet, MultiDropdownInputItem)
App.register_widget(InputPath, StrInputItem)
App.register_widget(OutputPath, StrInputItem)



from setuptools import setup, find_packages
import re


classifiers = [
    "Development Status :: 3 - Alpha",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
]


keywords = [
    'GUI',
    'CLI',
]


def get_version():
    with open("oneface/__init__.py") as f:
        for line in f.readlines():
            m = re.match("__version__ = '([^']+)'", line)
            if m:
                return m.group(1)
        raise IOError("Version information can not found.")


def get_long_description():
    return "See https://github.com/Nanguage/oneFace"


def get_install_requires():
    requirements = [
        "rich",
        "fire",
        "qtpy",
        "pyYAML",
    ]
    return requirements


requires_test = ['pytest', 'pytest-cov', 'flake8']
requires_doc = []
with open("docs/requirements.txt") as f:
    for line in f:
        p = line.strip()
        if p:
            requires_doc.append(p)
requires_dash = ['ansi2html', 'dash', 'visdcc']


setup(
    name='oneFace',
    author='Weize Xu',
    author_email='vet.xwz@gmail.com',
    version=get_version(),
    license='GPLv3',
    description='oneFace is a library for automatically generating multiple '
    'interfaces(CLI, GUI) from a callable Python object',
    long_description=get_long_description(),
    keywords=keywords,
    url='https://github.com/Nanguage/oneFace',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=classifiers,
    install_requires=get_install_requires(),
    extras_require={
        'test': requires_test,
        'doc': requires_doc,
        'dev': requires_test + requires_doc,
        'dash': requires_dash,
        'qt': ['pyqt6'],
        'pyqt5': ['pyqt5'],
        'pyqt6': ['pyqt6'],
        'pyside2': ['PySide2'],
        'pyside6': ['PySide6'],
        'all': ['pyqt6'] + requires_dash,
    },
    python_requires='>=3.7, <4',
)
